"""
Data collections mapping semantic units in the SEPA standard. Most of those
can be rendered to an XML representation.
"""

from datetime import datetime
from decimal import Decimal
from io import StringIO
from lxml.builder import ElementMaker
from lxml.etree import tostring, parse, XMLSchema
from pathlib import Path
import re


class PyPAINException(Exception):
    pass


class InvalidXMLException(PyPAINException):
    """Signifies XML that doesn't conform with the SEPA XML schema."""

    pass


class InvalidXMLValueException(PyPAINException):
    """Signifies values that don't conform with SEPA data types."""

    pass


class SEPAType:

    syntax = ""
    special_character_mapping = {
        "Ä": "Ae",
        "Ö": "Oe",
        "Ü": "Ue",
        "ä": "ae",
        "ö": "oe",
        "ü": "ue",
        "ß": "ss",
        "&": "+",
        "*": ".",
        "$": ".",
        "%": ".",
        "<": ".",
        ">": ".",
    }
    # Even though the SEPA scheme defines an UTF-8 encoding, only a small
    # subset is indeed valid. This may change.
    special_character_pattern = r"([^A-Za-z0-9\+|\?|/|\-|:|\(|\)|\.|,|\'| ])"
    special_character_fallback = "."

    def __init__(self, input_value, replace_special_chars=True):
        self.value = self.transform(input_value)
        if replace_special_chars:
            self.replace_special_characters()

        if not self.valid():
            raise InvalidXMLValueException(
                'Invalid value "{}" for type {}.'.format(
                    self.value, self.__class__.__name__
                )
            )

    @staticmethod
    def transform(input_value):
        """Transform input value to string representation."""
        return str(input_value)

    def replace_special_characters(self):
        for search, replace in self.special_character_mapping.items():
            self.value = self.value.replace(search, replace)
        self.value = re.sub(
            self.special_character_pattern,
            self.special_character_fallback,
            self.value,
        )

    def valid(self):
        """Return True if value is conform to the syntax of this type."""
        regex = re.compile(self.syntax)
        return regex.match(self.value) is not None

    @staticmethod
    def to_string(_, item):
        """Used by lxml's ETree to serialize values implicitly."""
        return item.value

    def __str__(self):
        return "{} <{}>".format(self.__class__.__name__, self.value)


class AnyBICIdentifier(SEPAType):
    """
    >>> AnyBICIdentifier('MARKDEFF').valid()
    True
    >>> AnyBICIdentifier('INVALID').valid()
    Traceback (most recent call last):
    pypain.exceptions.InvalidXMLValueException: ...
    """

    syntax = r"^[A-Z]{6,6}[A-Z2-9][A-NP-Z0-9]([A-Z0-9]{3,3}){0,1}$"


class BICIdentifier(SEPAType):
    """
    >>> BICIdentifier('MARKDEFF').valid()
    True
    >>> BICIdentifier('INVALID').valid()
    Traceback (most recent call last):
    pypain.exceptions.InvalidXMLValueException: ...
    """

    syntax = r"^[A-Z]{6,6}[A-Z2-9][A-NP-Z0-9]([A-Z0-9]{3,3}){0,1}$"


class CountryCode(SEPAType):
    """
    >>> CountryCode('DE').valid()
    True
    >>> CountryCode('A1').valid()
    Traceback (most recent call last):
    pypain.exceptions.InvalidXMLValueException: ...
    """

    syntax = r"^[A-Z]{2,2}$"


class ActiveOrHistoricCurrencyCode(SEPAType):
    """
    >>> ActiveOrHistoricCurrencyCode('USD').valid()
    True
    >>> ActiveOrHistoricCurrencyCode('DM').valid()
    Traceback (most recent call last):
    pypain.exceptions.InvalidXMLValueException: ...
    """

    syntax = r"^[A-Z]{3,3}$"


class ActiveOrHistoricCurrencyCodeEUR(SEPAType):
    """
    >>> ActiveOrHistoricCurrencyCodeEUR('EUR').valid()
    True
    >>> ActiveOrHistoricCurrencyCodeEUR('USD').valid()
    Traceback (most recent call last):
    pypain.exceptions.InvalidXMLValueException: ...
    """

    syntax = r"^EUR$"


class ActiveOrHistoricCurrencyAndAmountSEPA(SEPAType):
    """
    >>> ActiveOrHistoricCurrencyAndAmountSEPA('12345678901').valid()
    True
    >>> ActiveOrHistoricCurrencyAndAmountSEPA('0.12').valid()
    True
    >>> ActiveOrHistoricCurrencyAndAmountSEPA('0.1').valid()
    True
    >>> ActiveOrHistoricCurrencyAndAmountSEPA('-1').valid()
    Traceback (most recent call last):
    pypain.exceptions.InvalidXMLValueException: ...
    >>> ActiveOrHistoricCurrencyAndAmountSEPA('0.123').valid()
    Traceback (most recent call last):
    pypain.exceptions.InvalidXMLValueException: ...
    >>> ActiveOrHistoricCurrencyAndAmountSEPA(Decimal('1.1')).value
    '1.1'
    """

    syntax = r"^\d{1,11}(\.\d{1,2})?$"


class DecimalTime(SEPAType):
    """
    >>> DecimalTime('235959000').valid()
    True
    >>> DecimalTime('235959').valid()
    Traceback (most recent call last):
    pypain.exceptions.InvalidXMLValueException: ...
    """

    syntax = r"^[0-9]{9,9}$"


class IBAN2007Identifier(SEPAType):
    """
    >>> IBAN2007Identifier('KW81CBKU0000000000001234560101').valid()
    True
    >>> IBAN2007Identifier('NO8330001234567').valid()
    True
    >>> IBAN2007Identifier('DE12').valid()
    Traceback (most recent call last):
    pypain.exceptions.InvalidXMLValueException: ...
    """

    syntax = r"^[A-Z]{2,2}[0-9]{2,2}[a-zA-Z0-9]{1,30}$"


class Max1025Text(SEPAType):
    """
    >>> Max1025Text('A valid text.').valid()
    True
    >>> Max1025Text('A' * 1026).valid()
    Traceback (most recent call last):
    pypain.exceptions.InvalidXMLValueException: ...
    >>> Max1025Text('').valid()
    Traceback (most recent call last):
    pypain.exceptions.InvalidXMLValueException: ...
    """

    syntax = r"^.{1,1025}$"


class Max140Text(SEPAType):
    """
    >>> Max140Text('A valid text.').valid()
    True
    >>> Max140Text('A' * 141).valid()
    Traceback (most recent call last):
    pypain.exceptions.InvalidXMLValueException: ...
    >>> Max140Text('').valid()
    Traceback (most recent call last):
    pypain.exceptions.InvalidXMLValueException: ...
    """

    syntax = r"^.{1,140}$"


class Max15NumericText(SEPAType):
    """
    >>> Max15NumericText('1' * 15).valid()
    True
    >>> Max15NumericText('1' * 16).valid()
    Traceback (most recent call last):
    pypain.exceptions.InvalidXMLValueException: ...
    >>> Max15NumericText('').valid()
    Traceback (most recent call last):
    pypain.exceptions.InvalidXMLValueException: ...
    >>> Max15NumericText(42).value  # Number types are fine.
    '42'
    """

    syntax = r"^[0-9]{1,15}$"


class Max35Text(SEPAType):
    """
    >>> Max35Text('1' * 35).valid()
    True
    >>> Max35Text('1' * 36).valid()
    Traceback (most recent call last):
    pypain.exceptions.InvalidXMLValueException: ...
    >>> Max35Text('').valid()
    Traceback (most recent call last):
    pypain.exceptions.InvalidXMLValueException: ...
    """

    syntax = r"^.{1,35}$"


class Max70Text(SEPAType):
    """
    >>> Max70Text('1' * 70).valid()
    True
    >>> Max70Text('1' * 71).valid()
    Traceback (most recent call last):
    pypain.exceptions.InvalidXMLValueException: ...
    >>> Max70Text('').valid()
    Traceback (most recent call last):
    pypain.exceptions.InvalidXMLValueException: ...
    """

    syntax = r"^.{1,70}$"


class RestrictedIdentificationSEPA1(SEPAType):
    """
    >>> RestrictedIdentificationSEPA1('Identifier-1. ').valid()
    True
    >>> RestrictedIdentificationSEPA1('A' * 35).valid()
    True
    >>> RestrictedIdentificationSEPA1('A' * 36).valid()
    Traceback (most recent call last):
    pypain.exceptions.InvalidXMLValueException: ...
    """

    syntax = r"^([A-Za-z0-9]|[\+|\?|/|\-|:|\(|\)|\.|,|\'| ]){1,35}$"


class RestrictedIdentificationSEPA2(SEPAType):
    """
    >>> RestrictedIdentificationSEPA2('Identifier-2.').valid()
    True
    >>> RestrictedIdentificationSEPA2('A' * 35).valid()
    True
    >>> RestrictedIdentificationSEPA2('A' * 36).valid()
    Traceback (most recent call last):
    pypain.exceptions.InvalidXMLValueException: ...
    >>> RestrictedIdentificationSEPA2('with whitespace').valid()
    Traceback (most recent call last):
    pypain.exceptions.InvalidXMLValueException: ...
    """

    syntax = r"^([A-Za-z0-9]|[\+|\?|/|\-|:|\(|\)|\.|,|\']){1,35}$"


class RestrictedPersonIdentifierSEPA(SEPAType):
    """
    >>> RestrictedPersonIdentifierSEPA('DE71ZZZ00000111111').valid()
    True
    >>> RestrictedPersonIdentifierSEPA('Invalid').valid()
    Traceback (most recent call last):
    pypain.exceptions.InvalidXMLValueException: ...
    """

    syntax = (
        r"^[a-zA-Z]{2,2}[0-9]{2,2}([A-Za-z0-9]|[\+|\?|/|\-|:|\(|\)|\.|,|\']){3,3}"  # noqa
        r"([A-Za-z0-9]|[\+|\?|/|\-|:|\(|\)|\.|,|\']){1,28}$"
    )


class ExternalLocalInstrument1Code(SEPAType):
    """
    >>> ExternalLocalInstrument1Code('COR1').valid()
    True
    >>> ExternalLocalInstrument1Code('Invalid').valid()
    Traceback (most recent call last):
    pypain.exceptions.InvalidXMLValueException: ...
    """

    syntax = r"^(CORE|COR1|B2B)$"


class ExternalCategoryPurpose1Code(SEPAType):
    """
    >>> ExternalCategoryPurpose1Code('CASH').valid()
    True
    >>> ExternalCategoryPurpose1Code('Invalid').valid()
    Traceback (most recent call last):
    pypain.exceptions.InvalidXMLValueException: ...
    """

    syntax = (
        r"^(BONU|CASH|CBLK|CCRD|CORT|DCRD|DIVI|EPAY|FCOL|GOVT|HEDG|ICCP|IDCP|INTC|INTE|LOAN|OTHR|PENS|SALA|SECU|"  # noqa
        r"|SSBE|SUPP|TAXS|TRAD|TREA|VATX|WHLD)$"
    )


class ExternalPurpose1Code(SEPAType):
    """
    >>> ExternalPurpose1Code('CASH').valid()
    True
    >>> ExternalPurpose1Code('Invalid').valid()
    Traceback (most recent call last):
    pypain.exceptions.InvalidXMLValueException: ...
    """

    syntax = (
        r"^(CBLK|CDCB|CDCD|CDCS|CDDP|CDOC|CDQC|ETUP|FCOL|MTUP|ACCT|CASH|COLL|CSDB|DEPT|INTC|LIMA|NETT|AGRT|AREN|"  # noqa
        r"BEXP|BOCE|COMC|CPYR|GDDS|GDSV|GSCB|LICF|POPE|ROYA|SCVE|SUBS|SUPP|TRAD|CHAR|COMT|CLPR|DBTC|GOVI|HLRP|"  # noqa
        r"INPC|INSU|INTE|LBRI|LIFI|LOAN|LOAR|PENO|PPTI|RINP|TRFD|ADMG|ADVA|BLDM|CBFF|CCRD|CDBL|CFEE|COST|CPKC|"  # noqa
        r"DCRD|EDUC|FAND|FCPM|GOVT|ICCP|IDCP|IHRP|INSM|IVPT|MSVC|NOWS|OFEE|OTHR|PADD|PTSP|RCKE|RCPT|REBT|REFU|"  # noqa
        r"RENT|STDY|TBIL|TCSC|TELI|WEBI|ANNI|CAFI|CFDI|CMDT|DERI|DIVD|FREX|HEDG|INVS|PRME|SAVG|SECU|SEPI|TREA|"  # noqa
        r"ANTS|CVCF|DMEQ|DNTS|HLTC|HLTI|HSPC|ICRF|LTCF|MDCS|VIEW|ALLW|ALMY|BBSC|BECH|BENE|BONU|COMM|CSLP|GVEA|"  # noqa
        r"GVEB|GVEC|GVED|PAYR|PENS|PRCP|SALA|SSBE|AEMP|GFRP|GWLT|RHBS|ESTX|FWLV|GSTX|HSTX|INTX|NITX|PTXP|RDTX|"  # noqa
        r"TAXS|VATX|WHLD|TAXR|AIRB|BUSB|FERB|RLWY|TRPT|CBTV|ELEC|ENRG|GASB|NWCH|NWCM|OTLC|PHON|UBIL|WTER)$"  # noqa
    )


class SequenceType1Code(SEPAType):
    """
    >>> SequenceType1Code('FNAL').valid()
    True
    >>> SequenceType1Code('Invalid').valid()
    Traceback (most recent call last):
    pypain.exceptions.InvalidXMLValueException: ...
    """

    syntax = r"^(FRST|RCUR|OOFF|FNAL)$"


class ISODate(SEPAType):
    """
    >>> ISODate('2000-01-01').valid()
    True
    >>> ISODate('20000101').valid()
    True
    >>> ISODate('2000-50-01').valid()
    Traceback (most recent call last):
    pypain.exceptions.InvalidXMLValueException: ...
     >>> ISODate(date(2000, 1, 1)).value  # Date instances are fine.
     '2000-01-01'
    """

    syntax = r"^([0-9]{4})-?(1[0-2]|0[1-9])-?(3[0-1]|0[1-9]|[1-2][0-9])$"

    @staticmethod
    def transform(input_value):
        if hasattr(input_value, "isoformat"):
            return input_value.isoformat()
        return SEPAType.transform(input_value)


class ISODateTime(SEPAType):
    """
    >>> ISODateTime('2000-01-01T12:00:00Z').valid()
    True
    >>> ISODateTime('2000-01-01T12:00:00+01:00').valid()
    True
    >>> ISODateTime('2000-01-01T12:00:00').valid()
    True
    >>> ISODateTime('2000-01-01').valid()
    Traceback (most recent call last):
    pypain.exceptions.InvalidXMLValueException: ...
    >>> ISODateTime(datetime(2000, 1, 1, 12, 0, 0)).value
    '2000-01-01T12:00:00'
    """

    syntax = (
        r"^(-?(?:[1-9][0-9]*)?[0-9]{4})-(1[0-2]|0[1-9])-(3[0-1]|0[1-9]|[1-2][0-9])?"  # noqa
        r"T(2[0-3]|[0-1][0-9]):([0-5][0-9]):([0-5][0-9])(\.[0-9]+)??(Z|[+-](?:2[0-3]|[0-1][0-9]):[0-5][0-9])?$"  # noqa
    )

    @staticmethod
    def transform(input_value):
        if hasattr(input_value, "isoformat"):
            return input_value.isoformat()
        return SEPAType.transform(input_value)


class DecimalNumber(SEPAType):
    """
    >>> DecimalNumber('1').valid()
    True
    >>> DecimalNumber('{}.{}'.format('1' * 18, '1' * 17)).valid()
    True
    >>> DecimalNumber('1' * 19).valid()
    Traceback (most recent call last):
    pypain.exceptions.InvalidXMLValueException: ...
    >>> DecimalNumber('1.001').value  # Numbers are truncated to two decimals.
    '1.00'
    """

    syntax = r"^\d{1,18}(\.\d{1,17})?$"

    @staticmethod
    def transform(input_value):
        return "{:.2f}".format(Decimal(input_value))


class BoolString(SEPAType):
    """
    >>> BoolString('true').valid()
    True
    >>> BoolString('invalid').valid()
    Traceback (most recent call last):
    pypain.exceptions.InvalidXMLValueException: ...
    """

    syntax = r"^(false|true)$"

    @staticmethod
    def transform(input_value):
        return str(input_value).lower()


class NullElementMaker(ElementMaker):

    def __call__(self, *args, **kwargs):
        filtered_args = [arg for arg in args if arg is not None]
        return ElementMaker.__call__(self, *filtered_args, **kwargs)


E = NullElementMaker(
    typemap={SEPAType: SEPAType.to_string},
    nsmap={
        None: "urn:iso:std:iso:20022:tech:xsd:pain.008.003.02",
        "xsi": "http://www.w3.org/2001/XMLSchema-instance",
    },
)

D = NullElementMaker(
    typemap={SEPAType: SEPAType.to_string},
    nsmap={
        None: "urn:iso:std:iso:20022:tech:xsd:pain.001.001.03",
        "xsi": "http://www.w3.org/2001/XMLSchema-instance",
    },
)


class XMLProducer:
    """Models a data class that is able to render to an XML representation."""

    xml_schema_path = None

    def to_xml(self):
        """Return lxml tree structure."""
        raise NotImplementedError(
            "Implement method `to_xml` to generate XML structure."
        )

    def valid(self):
        try:
            self.to_xml()
        except InvalidXMLValueException:
            return False
        return True


class SEPAMessage(XMLProducer):

    xml_schema_path = None

    """Models a data class that is able to render to an XML representation."""

    def to_string(self, validate_xml=True):
        """Return rendered XML as string."""
        xml_string = tostring(
            self.to_xml(), xml_declaration=False, encoding="unicode"
        )
        if validate_xml:
            if not self.valid_xml(xml_string):
                raise InvalidXMLException
        return xml_string

    def to_file(self, file, validate_xml=True):
        """Write rendered XML to file-like object `file`."""
        xml_string = tostring(
            self.to_xml(), xml_declaration=False, encoding="unicode"
        )
        if validate_xml:
            if not self.valid_xml(xml_string):
                raise InvalidXMLException
        file.write(xml_string)

    def valid_xml(self, xml_string):
        """Return True if `xml_string` conforms to SEPA XML schema."""
        if self.xml_schema_path is None:
            raise NotImplementedError(
                "XSD validation failed: path to schema is not set."
            )
        parsed_xml = parse(StringIO(xml_string))
        xsd_scheme = XMLSchema(parse(self.xml_schema_path))
        return xsd_scheme.validate(parsed_xml)


class Creditor:
    """Data collection of creditor information in a direct debit."""

    def __init__(self, name, iban, bic, ci, currency=None, ultimate_name=None):
        self.name = name
        self.iban = iban
        self.bic = bic
        self.ci = ci
        self.currency = currency
        self.ultimate_name = ultimate_name

    def __str__(self):
        return "Creditor <{}>".format(self.name)


class Debitor:
    """Data collection of debitor information in a bank wire."""

    def __init__(self, name, iban, bic, currency=None, ultimate_name=None):
        self.name = name
        self.iban = iban
        self.bic = bic
        self.currency = currency
        self.ultimate_name = ultimate_name

    def __str__(self):
        return "Debitor <{}>".format(self.name)


class Transaction(XMLProducer):

    def __init__(
        self,
        debtor_name,
        debtor_iban,
        debtor_bic,
        amount,
        purpose,
        mandate_id,
        mandate_date,
        reason,
        end_to_end_id=None,
        currency="EUR",
        ultimate_name=None,
        old_mandate_id=None,
        old_creditor_name=None,
        old_creditor_id=None,
        old_debtor_iban=None,
        new_debtor_agent=False,
    ):
        self._debtor_name = debtor_name
        self._debtor_iban = debtor_iban
        self._debtor_bic = debtor_bic
        self.amount = amount
        self._purpose = purpose
        self._mandate_id = mandate_id
        self._mandate_date = mandate_date
        self._mandate_changed = (
            old_mandate_id is not None or
            old_creditor_name is not None or
            old_creditor_id is not None or
            old_debtor_iban is not None or
            new_debtor_agent
        )
        self._reason = reason
        self._end_to_end_id = end_to_end_id
        self._currency = currency
        self._ultimate_name = ultimate_name
        self._old_mandate_id = old_mandate_id
        self._old_creditor_name = old_creditor_name
        self._old_creditor_id = old_creditor_id
        self._old_debtor_iban = old_debtor_iban
        self._new_debtor_agent = new_debtor_agent

    def to_xml(self):
        """Return lxml tree structure."""
        # Individual transaction
        root = E.DrctDbtTxInf(
            # Reference a single payment instruction
            E.PmtId(
                # Unambiguous reference of the submitter of a direct debit.
                E.EndToEndId(
                    RestrictedIdentificationSEPA1(self._end_to_end_id)
                    if self._end_to_end_id is not None
                    else "NOTPROVIDED"
                )
            ),
            # Amount of money to be moved
            E.InstdAmt(
                ActiveOrHistoricCurrencyAndAmountSEPA(self.amount),
                Ccy=self._currency,
            ),
            # Information specific to the direct debit mandate
            E.DrctDbtTx(
                # Further mandate information
                E.MndtRltdInf(
                    E.MndtId(RestrictedIdentificationSEPA2(self._mandate_id)),
                    # Date on which the mandate was signed
                    E.DtOfSgntr(ISODate(self._mandate_date)),
                    *[
                        # Signifier of a mandate change
                        E.AmdmntInd("true"),
                        # Further mandate change information
                        E.AmdmntInfDtls(
                            # Original mandate reference
                            E.OrgnlMndtId(
                                RestrictedIdentificationSEPA2(
                                    self._old_mandate_id
                                )
                            )
                            if self._old_mandate_id is not None
                            else None,
                            # Original creditor scheme information
                            E.OrgnlCdtrSchmeId(
                                # Name of former creditor
                                E.Nm(Max70Text(self._old_creditor_name))
                                if self._old_creditor_name is not None
                                else None,
                                *[
                                    # Identification of an individual
                                    E.Id(
                                        E.PrvtId(
                                            # Other unambiguous identifier
                                            E.Othr(
                                                # Original CI of creditor
                                                E.Id(
                                                    RestrictedPersonIdentifierSEPA(  # noqa
                                                        self._old_creditor_id
                                                    )
                                                ),
                                                E.SchmeNm(
                                                    # Name in text
                                                    E.Prtry("SEPA")
                                                ),
                                            )
                                        )
                                    )
                                ]
                                if self._old_creditor_id is not None
                                else []
                            )
                            if self._old_creditor_name is not None or
                                self._old_creditor_id is not None  # noqa
                            else None,
                            # Original debtor account
                            E.OrgnlDbtrAcct(
                                # Identification of account
                                E.Id(
                                    # International Bank Account Number (IBAN)
                                    E.IBAN(
                                        IBAN2007Identifier(
                                            self._old_debtor_iban
                                        )
                                    )
                                )
                            )
                            if self._old_debtor_iban is not None
                            else None,
                            # Original debtor's agent
                            E.OrgnlDbtrAgt(
                                # Identification of debtor's agent
                                E.FinInstnId(
                                    E.Othr(
                                        E.Id(
                                            "SMNDA"
                                        )  # Same Mandate, New Debtor Agent
                                    )
                                )
                            )
                            if self._new_debtor_agent
                            else None,
                        ),
                    ]
                    if self._mandate_changed
                    else []
                )
            ),
            # Financial institution of debtor
            E.DbtrAgt(
                # Unique identification of institution
                E.FinInstnId(
                    # Business Identifier Code (SWIFT-Code)
                    E.BIC(BICIdentifier(self._debtor_bic))
                    if self._debtor_bic is not None
                    else None,
                    # Other identification of financial institutions
                    E.Othr(E.Id("NOTPROVIDED"))
                    if self._debtor_bic is None
                    else None,
                )
            ),
            # Information about the debtor
            E.Dbtr(
                # Name
                E.Nm(Max70Text(self._debtor_name))
            ),
            # Bank account of debtor
            E.DbtrAcct(
                # Identification of debtor's bank account
                E.Id(
                    # International Bank Account Number (IBAN)
                    E.IBAN(IBAN2007Identifier(self._debtor_iban))
                )
            ),
            # Debtor, if not bank account owner himself
            E.UltmtDbtr(
                # Name
                E.Nm(Max70Text(self._ultimate_name))
            )
            if self._ultimate_name is not None
            else None,
            # Type of payment
            E.Purp(
                # Type in coded format
                E.Cd(ExternalPurpose1Code(self._purpose))
            )
            if self._purpose is not None
            else None,
            # Remittance information
            E.RmtInf(
                # Unstructured information
                E.Ustrd(Max140Text(self._reason))
            )
            if self._reason is not None
            else None,
        )
        return root

    def __str__(self):
        return "Transaction <{amount} from {name}>".format(
            amount=self.amount, name=self._debtor_name
        )


class Payment(XMLProducer):
    """
    """

    def __init__(
        self,
        creditor,
        payment_id,
        direct_debit_type,
        sequence_type,
        purpose,
        payment_date,
        batch_booking=True,
    ):
        self._creditor = creditor
        self._payment_id = payment_id
        self._direct_debit_type = direct_debit_type
        self._sequence_type = sequence_type
        self._purpose = purpose
        self._payment_date = payment_date
        self._batch_booking = batch_booking
        self._transactions = []

    def add_transaction(
        self,
        debtor_name,
        debtor_iban,
        debtor_bic,
        amount,
        purpose,
        mandate_id,
        mandate_date,
        reason,
        end_to_end_id=None,
        currency="EUR",
        ultimate_name=None,
        old_mandate_id=None,
        old_creditor_name=None,
        old_creditor_id=None,
        old_debtor_iban=None,
        new_debtor_agent=False,
    ):
        transaction = Transaction(
            debtor_name=debtor_name,
            debtor_iban=debtor_iban,
            debtor_bic=debtor_bic,
            amount=amount,
            purpose=purpose,
            mandate_id=mandate_id,
            mandate_date=mandate_date,
            reason=reason,
            end_to_end_id=end_to_end_id,
            currency=currency,
            ultimate_name=ultimate_name,
            old_mandate_id=old_mandate_id,
            old_creditor_name=old_creditor_name,
            old_creditor_id=old_creditor_id,
            old_debtor_iban=old_debtor_iban,
            new_debtor_agent=new_debtor_agent,
        )
        self._transactions.append(transaction)
        return transaction

    def count_transactions(self):
        """Return number of associated transactions."""
        return len(self._transactions)

    def transactions_sum(self):
        """Returns sum of all transaction amounts in this payment."""
        return sum(transaction.amount for transaction in self._transactions)

    def to_xml(self):
        """Return lxml tree structure."""
        root = E.PmtInf(
            # Unique reference assigned to payment
            E.PmtInfId(RestrictedIdentificationSEPA1(self._payment_id)),
            # Means of payment, here: Direct Debit
            E.PmtMtd("DD"),
            E.BtchBookg(BoolString(self._batch_booking)),
            # Number of included transactions
            E.NbOfTxs(Max15NumericText(len(self._transactions))),
            # Sum of all included transactions with two decimal places
            E.CtrlSum(DecimalNumber(self.transactions_sum())),
            # Type of transaction
            E.PmtTpInf(
                # Agreement under which the transactions should be processed
                E.SvcLvl(
                    # In coded format
                    E.Cd("SEPA")
                ),
                # Type of direct debit
                E.LclInstrm(
                    # In coded format
                    E.Cd(ExternalLocalInstrument1Code(self._direct_debit_type))
                ),
                E.SeqTp(SequenceType1Code(self._sequence_type)),
                # Type of payment
                E.CtgyPurp(
                    # In coded format
                    E.Cd(ExternalCategoryPurpose1Code(self._purpose))
                )
                if self._purpose is not None
                else None,
            ),
            # Requested collection date
            E.ReqdColltnDt(ISODate(self._payment_date)),
            # Creditor Information
            E.Cdtr(
                # Name
                E.Nm(Max70Text(self._creditor.name))
            ),
            # Bank account of creditor
            E.CdtrAcct(
                # Identification of bank account
                E.Id(
                    # International Bank Account Number (IBAN)
                    E.IBAN(IBAN2007Identifier(self._creditor.iban))
                ),
                # Currency of bank account
                E.Ccy(ActiveOrHistoricCurrencyCode(self._creditor.currency))
                if self._creditor.currency is not None
                else None,
            ),
            # Financial institution of creditor
            E.CdtrAgt(
                # Identification of the institution
                E.FinInstnId(
                    # Business Identifier Code (SWIFT-Code)
                    E.BIC(BICIdentifier(self._creditor.bic))
                    if self._creditor.bic is not None
                    else None,
                    # Other means of identification
                    E.Othr(E.Id("NOTPROVIDED"))
                    if self._creditor.bic is None
                    else None,
                )
            ),
            # Name of ultimate creditor, if not the creditor above.
            E.UltmtCdtr(
                # Name
                E.Nm(Max70Text(self._creditor.ultimate_name))
            )
            if self._creditor.ultimate_name is not None
            else None,
            E.ChrgBr("SLEV"),  # Both parties bear their own charges
            # Credit party that signs the mandate
            E.CdtrSchmeId(
                # Identification of an organisation or person
                E.Id(
                    # Unique identification of a person
                    E.PrvtId(
                        # Other types of identification
                        E.Othr(
                            # Creditor Identification (CI)
                            E.Id(
                                RestrictedPersonIdentifierSEPA(
                                    self._creditor.ci
                                )
                            ),
                            # Name of identification scheme
                            E.SchmeNm(
                                # Name in free text
                                E.Prtry("SEPA")
                            ),
                        )
                    )
                )
            ),
            *[transaction.to_xml() for transaction in self._transactions]
        )
        return root

    def __iter__(self):
        """Iterate over transactions."""
        yield from self._transactions

    def __getitem__(self, key):
        """Access transaction via index."""
        return self._transactions[key]

    def __str__(self):
        return "Payment <{}>".format(self._payment_id)


class CreditTransaction(XMLProducer):

    def __init__(
        self,
        creditor_name,
        creditor_iban,
        creditor_bic,
        amount,
        reason,
        currency="EUR",
    ):
        self._creditor_name = creditor_name
        self._creditor_iban = creditor_iban
        self._creditor_bic = creditor_bic
        self.amount = amount
        self._reason = reason
        self._currency = currency
        self._end_to_end_id = None

    def to_xml(self):
        """Return lxml tree structure."""
        # Individual transaction
        root = D.CdtTrfTxInf(
            # Reference a single payment instruction
            D.PmtId(
                # Unambiguous reference of the submitter of a direct debit.
                D.EndToEndId(
                    RestrictedIdentificationSEPA1(self._end_to_end_id)
                    if self._end_to_end_id is not None
                    else "NOTPROVIDED"
                )
            ),
            # Amount of money to be moved
            D.Amt(
                D.InstdAmt(
                    ActiveOrHistoricCurrencyAndAmountSEPA(self.amount),
                    Ccy=self._currency,
                )
            ),
            # Information about the debtor
            D.Cdtr(
                # Name
                D.Nm(Max70Text(self._creditor_name))
            ),
            # Bank account of debtor
            D.CdtrAcct(
                # Identification of debtor's bank account
                D.Id(
                    # International Bank Account Number (IBAN)
                    D.IBAN(IBAN2007Identifier(self._creditor_iban))
                )
            ),
            # Remittance information
            D.RmtInf(
                # Unstructured information
                D.Ustrd(Max140Text(self._reason))
            )
            if self._reason is not None
            else None,
        )
        return root

    def __str__(self):
        return "Credit Transaction <{amount} from {name}>".format(
            amount=self.amount, name=self._creditor_name
        )


class Transfer(XMLProducer):
    """
    """

    def __init__(
        self,
        debitor,
        payment_id,
        payment_date,
    ):
        self._debitor = debitor
        self._payment_id = payment_id
        self._payment_date = payment_date
        self._transactions = []

    def add_transaction(
        self,
        creditor_name,
        creditor_iban,
        creditor_bic,
        amount,
        reason,
        currency="EUR",
    ):
        transaction = CreditTransaction(
            creditor_name=creditor_name,
            creditor_iban=creditor_iban,
            creditor_bic=creditor_bic,
            amount=amount,
            reason=reason,
            currency=currency,
        )
        self._transactions.append(transaction)
        return transaction

    def count_transactions(self):
        """Return number of associated transactions."""
        return len(self._transactions)

    def transactions_sum(self):
        """Returns sum of all transaction amounts in this payment."""
        return sum(transaction.amount for transaction in self._transactions)

    def to_xml(self):
        """Return lxml tree structure."""
        root = D.PmtInf(
            # Unique reference assigned to payment
            D.PmtInfId(RestrictedIdentificationSEPA1(self._payment_id)),
            # Means of payment, here: Direct Debit
            D.PmtMtd("TRF"),
            # Number of included transactions
            D.NbOfTxs(Max15NumericText(len(self._transactions))),
            # Sum of all included transactions with two decimal places
            D.CtrlSum(DecimalNumber(self.transactions_sum())),
            # Requested collection date
            D.ReqdExctnDt(ISODate(self._payment_date)),
            D.Dbtr(
                # Name
                D.Nm(Max70Text(self._debitor.name))
            ),
            # Bank account of creditor
            D.DbtrAcct(
                # Identification of bank account
                D.Id(
                    # International Bank Account Number (IBAN)
                    D.IBAN(IBAN2007Identifier(self._debitor.iban))
                ),
                # Currency of bank account
                D.Ccy(ActiveOrHistoricCurrencyCode(self._debitor.currency))
                if self._debitor.currency is not None
                else None,
            ),
            # Financial institution of creditor
            D.DbtrAgt(
                # Identification of the institution
                D.FinInstnId(
                    # Business Identifier Code (SWIFT-Code)
                    D.BIC(BICIdentifier(self._debitor.bic))
                    if self._debitor.bic is not None
                    else None,
                    # Other means of identification
                    D.Othr(E.Id("NOTPROVIDED"))
                    if self._debitor.bic is None
                    else None,
                )
            ),
            # Name of ultimate creditor, if not the creditor above.
            D.UltmtCdtr(
                # Name
                D.Nm(Max70Text(self._debitor.ultimate_name))
            )
            if self._debitor.ultimate_name is not None
            else None,
            D.ChrgBr("SLEV"),  # Both parties bear their own charges
            # Credit party that signs the mandate
            *[transaction.to_xml() for transaction in self._transactions]
        )
        return root

    def __iter__(self):
        """Iterate over transactions."""
        yield from self._transactions

    def __getitem__(self, key):
        """Access transaction via index."""
        return self._transactions[key]

    def __str__(self):
        return "Transfer <{}>".format(self._payment_id)


class BankWire(SEPAMessage):
    """SEPA bank wire message, containing payments and transactions."""

    xml_schema_path = str(
        Path(__file__).absolute().parent / "pain.001.001.03.xsd"
    )

    def __init__(
        self,
        message_id,
        debitor_name,
        debitor_iban,
        debitor_bic,
        debitor_currency=None,
        debitor_ultimate_name=None,
    ):
        self.message_id = message_id
        self._debitor = Debitor(
            name=debitor_name,
            iban=debitor_iban,
            bic=debitor_bic,
            currency=debitor_currency,
            ultimate_name=debitor_ultimate_name,
        )
        self._transfers = []

    def add_transfer(
        self,
        payment_id,
        payment_date,
        batch_booking=True,
    ):
        transfer = Transfer(
            debitor=self._debitor,
            payment_id=payment_id,
            payment_date=payment_date,
        )
        self._transfers.append(transfer)
        return transfer

    def count_transfers(self):
        return len(self._transfers)

    def transactions_sum(self):
        return sum(
            [transfer.transactions_sum() for transfer in self._transfers]
        )

    def count_transactions(self):
        """Return number of all transaction of this bank transfer."""
        return sum(
            [transfer.count_transactions() for transfer in self._transfers]
        )

    def to_xml(self, check_xsd=False):
        """Return lxml tree structure."""
        root = D.Document(
            # Customer Credit Transfer Initiation
            D.CstmrCdtTrfInitn(
                # Characteristics shared by all transactions in this message
                D.GrpHdr(
                    # End-to-end reference of this message
                    D.MsgId(RestrictedIdentificationSEPA1(self.message_id)),
                    # Date and time of file generation
                    D.CreDtTm(ISODateTime(datetime.now().isoformat())),
                    # Number of individual transaction in this file
                    D.NbOfTxs(Max15NumericText(self.count_transactions())),
                    D.CtrlSum(DecimalNumber(self.transactions_sum())),
                    # Initiating party
                    D.InitgPty(
                        # Name
                        D.Nm(Max70Text(self._debitor.name))
                    ),
                ),
                *[transfer.to_xml() for transfer in self._transfers]
            ),
            **{
                "{http://www.w3.org/2001/XMLSchema-instance}schemaLocation": (
                    "urn:iso:std:iso:20022:tech:xsd:pain.001.001.03 "
                    "pain.001.001.03.xsd")
            }
        )
        return root

    def __iter__(self):
        """Iterate over assigned payments."""
        yield from self._transfers

    def __getitem__(self, key):
        """Access payments via index. """
        return self._transfers[key]

    def __str__(self):
        return " <{}>".format(self.message_id)


class DirectDebit(SEPAMessage):
    """SEPA Direct Debit message, containing payments and transactions."""

    xml_schema_path = str(
        Path(__file__).absolute().parent / "pain.008.003.02.xsd"
    )

    def __init__(
        self,
        message_id,
        creditor_name,
        creditor_ci,
        creditor_iban,
        creditor_bic=None,
        creditor_currency=None,
        creditor_ultimate_name=None,
    ):
        """
        Initialize a SEPA direct debit process.
        :param message_id: unique, point-to-point reference for this process
        :param creditor_name: name of creditor
        :param creditor_ci: Creditor Identifier (CI)
        :param creditor_iban: creditor's bank account IBAN
        :param creditor_bic: creditor's bank account BIC
        :param creditor_currency: creditor's bank account currency
        :param creditor_ultimate_name: creditor reference party
        """
        self.message_id = message_id
        self._creditor = Creditor(
            name=creditor_name,
            iban=creditor_iban,
            bic=creditor_bic,
            ci=creditor_ci,
            currency=creditor_currency,
            ultimate_name=creditor_ultimate_name,
        )
        self._payments = []

    def add_payment(
        self,
        payment_id,
        direct_debit_type,
        sequence_type,
        purpose,
        payment_date,
        batch_booking=True,
    ):
        payment = Payment(
            creditor=self._creditor,
            payment_id=payment_id,
            direct_debit_type=direct_debit_type,
            sequence_type=sequence_type,
            purpose=purpose,
            payment_date=payment_date,
            batch_booking=batch_booking,
        )
        self._payments.append(payment)
        return payment

    def count_payments(self):
        """Return number of assigned payments."""
        return len(self._payments)

    def transactions_sum(self):
        return sum([payment.transactions_sum() for payment in self._payments])

    def count_transactions(self):
        """Return number of all transaction of this Direct Debit."""
        return sum(
            [payment.count_transactions() for payment in self._payments]
        )

    def to_xml(self, check_xsd=False):
        """Return lxml tree structure."""
        root = E.Document(
            # Customer Direct Debit Initiation
            E.CstmrDrctDbtInitn(
                # Characteristics shared by all transactions in this message
                E.GrpHdr(
                    # End-to-end reference of this message
                    E.MsgId(RestrictedIdentificationSEPA1(self.message_id)),
                    # Date and time of file generation
                    E.CreDtTm(ISODateTime(datetime.now().isoformat())),
                    # Number of individual transaction in this file
                    E.NbOfTxs(Max15NumericText(self.count_transactions())),
                    E.CtrlSum(DecimalNumber(self.transactions_sum())),
                    # Initiating party
                    E.InitgPty(
                        # Name
                        E.Nm(Max70Text(self._creditor.name))
                    ),
                ),
                *[payment.to_xml() for payment in self._payments]
            ),
            **{
                "{http://www.w3.org/2001/XMLSchema-instance}schemaLocation": (
                    "urn:iso:std:iso:20022:tech:xsd:pain.008.003.02 "
                    "pain.008.003.02.xsd")
            }
        )
        return root

    def __iter__(self):
        """Iterate over assigned payments."""
        yield from self._payments

    def __getitem__(self, key):
        """Access payments via index. """
        return self._payments[key]

    def __str__(self):
        return "Direct Debit <{}>".format(self.message_id)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from ..assignment import get_assignment_mail_data
from ..electricity import get_energyvalue_mail_data
from ..log import user_data_log, log_with_user
from .base import date_time, get_selected_year, route_url
from .letter import render_pdf, send_mail
from .protocol import format_markdown
from babel.dates import format_datetime
from PyPDF2 import PdfFileWriter, PdfFileReader
from bs4 import BeautifulSoup
from datetime import datetime
from io import BytesIO
from pyramid.response import Response, FileIter
from pyramid.security import NO_PERMISSION_REQUIRED
from pyramid.view import view_config
from sqlalchemy import func, or_
from sw.allotmentclub import ExternalRecipient, SentMessageInfo
from sw.allotmentclub import Member, Message, Attachment, User
from .external import filter_external_recipients
import base64
import collections
import dateutil.parser
import pybars
import pytz
import re
import sw.allotmentclub.browser.base


def append_pdf(input, output):
    [output.addPage(input.getPage(page_num)) for page_num in
        range(input.numPages)]


def get_recipient(value, request=None):
    recipients = Message.get(value).members + Message.get(value).externals
    if recipients:
        if (len(recipients)) > 1:
            return 'Mehrere Empfänger'
        recipient = recipients[0]
        if recipient.lastname == 'Verein':
            return 'Alle Mitglieder'
        return '{}, {}'.format(recipient.lastname, recipient.firstname)


def print_or_sent_date(value, request=None):
    message = Message.get(value)
    if message.sent:
        return date_time(message.sent)
    if message.printed:
        return date_time(message.printed)


def print_or_sent_type(value, request=None):
    message = Message.get(value)
    if message.sent and message.printed:
        return 'Brief und E-Mail'
    if message.sent:
        return 'E-Mail'
    if message.printed:
        return 'Brief'


class MailQuery(sw.allotmentclub.browser.base.Query):

    formatters = {
        'Gesendet': print_or_sent_date,
        'Via': print_or_sent_type,
        'Empfänger': get_recipient,
    }

    data_class = {
        'Betreff': 'expand'
    }
    data_hide = {
        'Gesendet': 'phone,tablet',
        'Via': 'phone,tablet',
        'Empfänger': 'phone,tablet',
        'Von': 'phone,tablet',
    }

    def select(self):
        return (
            self.db.query(
                Message.id.label('#'),
                Message.id.label('Empfänger'),
                Message.subject.label('Betreff'),
                User.username.label('Von'),
                Message.id.label('Gesendet'),
                Message.id.label('Via'))
            .select_from(Message)
            .join(User))


class InboxQuery(MailQuery):

    formatters = {
        'Empfangen': print_or_sent_date,
        'Absender': get_recipient,
    }
    data_class = {
        'Betreff': 'expand'
    }
    data_hide = {
    }

    def select(self):
        return (
            self.db.query(
                Message.id.label('#'),
                Message.subject.label('Betreff'),
                Message.id.label('Absender'),
                Message.id.label('Empfangen'))
            .select_from(Message)
            .join(User)
            .filter(Message.inbound.is_(True))
            .filter(Message.accounting_year == get_selected_year()))


class DraftsQuery(MailQuery):

    def select(self):
        query = super(DraftsQuery, self).select()
        return (
            query
            .filter(Message.inbound.is_(False))
            .filter(Message.sent.is_(None))
            .filter(Message.printed.is_(None)))


class SentQuery(MailQuery):

    def select(self):
        query = super(SentQuery, self).select()
        return (
            query
            .filter(Message.inbound.is_(False))
            .filter(or_(
                Message.sent.isnot(None),
                Message.printed.isnot(None)))
            .filter(Message.accounting_year == get_selected_year()))


@view_config(route_name='mail_list_drafts', renderer='json', permission='view')
class MailListDraftsView(sw.allotmentclub.browser.base.TableView):
    """ Liste aller Entwürfe."""

    query_class = DraftsQuery
    available_actions = [
        dict(url='mail_add', btn_class='btn-success', icon='fa fa-plus',
             title='Neu'),
        dict(url='mail_edit', btn_class='btn-success', icon='fa fa-pencil',
             title='Bearbeiten'),
        dict(url='mail_duplicate', btn_class='btn-success',
             icon='fa fa-copy', title='Duplizieren'),
        dict(url='mail_preview', btn_class='btn-success',
             icon='fa fa-print', title='Vorschau'),
        dict(url='mail_send', btn_class='btn-success',
             icon='fa fa-envelope', title='Versenden'),
        dict(url='mail_delete', btn_class='btn-danger',
             icon='glyphicon glyphicon-trash', title='Löschen')]


@view_config(route_name='mail_list_sent', renderer='json', permission='view')
class MailListSentView(sw.allotmentclub.browser.base.TableView):

    query_class = SentQuery
    default_order_by = 'sent'
    year_selection = True
    start_year = 2015
    available_actions = [
        dict(url='mail_status', btn_class='btn-success', icon='fa fa-truck',
             title='Zustell-Status'),
        dict(url='mail_edit', btn_class='btn-success', icon='fa fa-eye',
             title='Anzeigen'),
        dict(url='mail_preview', btn_class='btn-success',
             icon='fa fa-print', title='Vorschau'),
        dict(url='mail_send', btn_class='btn-success',
             icon='fa fa-envelope', title='Erneut Drucken'),
        dict(url='mail_duplicate', btn_class='btn-success',
             icon='fa fa-copy', title='Duplizieren')]


@view_config(route_name='mail_list_inbox', renderer='json', permission='view')
class MailListInboxView(sw.allotmentclub.browser.base.TableView):

    query_class = InboxQuery
    default_order_by = 'sent'
    year_selection = True
    start_year = 2017
    available_actions = [
        dict(url='mail_edit', btn_class='btn-success', icon='fa fa-eye',
             title='Anzeigen'),
        dict(url='mail_reply', btn_class='btn-success', icon='fa fa-reply',
             title='Antworten'),
        dict(url='mail_edit', btn_class='btn-success', icon='fa fa-pencil',
             title='Anzeigen')]


class MailStatusQuery(sw.allotmentclub.browser.base.Query):

    data_class = {
        'Empfänger': 'expand'
    }

    def select(self):
        return (
            self.db.query(
                SentMessageInfo.id.label('#'),
                SentMessageInfo.address.label('Empfänger'),
                func.coalesce(
                    SentMessageInfo.status, 'Versendet').label('Status'))
            .select_from(SentMessageInfo)
            .join(Message)
            .filter(SentMessageInfo.message == self.context))


@view_config(route_name='mail_status', renderer='json', permission='view')
class MailStatusView(sw.allotmentclub.browser.base.TableView):
    """Liste aller Empfänger-Status."""

    query_class = MailStatusQuery
    default_order_by = 'address'
    year_selection = False


@view_config(route_name='mail_preview', renderer='json', permission='view')
class MailPreviewView(sw.allotmentclub.browser.base.View):

    preview = True
    additional = {'Energieabrechnung': get_energyvalue_mail_data,
                  'Fehlende Arbeitsstunden': get_assignment_mail_data}

    def add_attachments(self, pdf, output):
        if self.context.attachments:
            for attachment in self.context.attachments:
                if attachment.mimetype in ('application/pdf',):
                    if attachment.white_page_before:
                        output.addBlankPage()
                    pdf = PdfFileReader(BytesIO(attachment.data), strict=False)
                    append_pdf(pdf, output)
        return pdf, output

    @property
    def recipients(self):
        recipients = self.context.members + self.context.externals
        organization_id = self.request.user.organization_id
        if not recipients:
            return [None]
        members = self.context.members
        if members and members[0].lastname == 'Verein':
            return (
                Member.query()
                .filter(Member.leaving_year.is_(None))
                .filter(Member.organization_id == organization_id)
                .filter(Member.get_post.is_(True)).all())
        return recipients

    def recipients_data(self, recipient):
        if recipient is None:
            return dict()
        return dict(deflection=recipient.deflection,
                    appellation=recipient.appellation,
                    title=recipient.title,
                    organization=recipient.organization or '',
                    firstname=recipient.firstname,
                    lastname=recipient.lastname,
                    phone=getattr(recipient, 'phone', ''),
                    mobile=getattr(recipient, 'mobile', ''))

    def update(self):
        output = PdfFileWriter()
        if self.recipients == [None] or self.preview:
            recipients = self.recipients
        else:
            recipients = [r for r in self.recipients if r and not r.email]
        subject = self.context.subject or ''
        compiler = pybars.Compiler()
        compiled_body = compiler.compile(self.context.body)
        for recipient in recipients:
            address = recipient.address if recipient else None
            data = self.recipients_data(recipient)
            if subject in self.additional.keys():
                for nsubject, additional_data in self.additional[subject](
                        recipient, self.context.accounting_year):
                    additional_data.update(data)
                    body = format_markdown(compiled_body(additional_data))
                    output = self.add_page(address, nsubject, body, output)
            else:
                body = format_markdown(compiled_body(data))
                output = self.add_page(address, subject, body, output)
        result = BytesIO()
        output.write(result)
        result.seek(0)
        response = self.request.response
        response.set_cookie('fileDownload', value='true')
        response.content_type = 'application/pdf'
        response.content_disposition = 'attachment; filename=Brief.pdf'
        response.app_iter = FileIter(result)
        self.result = response

    def add_page(self, address, subject, body, output):
        sent = self.context.printed or self.context.sent
        pdf = render_pdf(address, subject, body, self.context.user,
                         date=sent, force_from=False)
        pdf = PdfFileReader(pdf, strict=False)
        append_pdf(pdf, output)
        pdf, output = self.add_attachments(pdf, output)
        numpages = pdf.numPages
        if numpages % 2 == 1:
            output.addBlankPage()
        return output


@view_config(route_name='mail_print', renderer='json', permission='view')
class MailPrintView(MailPreviewView):

    preview = False

    def update(self):
        if not self.context.printed:
            self.context.printed = datetime.now()
        return super(MailPrintView, self).update()


@view_config(route_name='mail_send', renderer='json', permission='view')
class MailSendView(MailPrintView):

    def update(self):
        if self.recipients == [None]:
            return
        recipients = []
        if not self.context.sent:
            recipients = [r for r in self.recipients if r.email]
            subject = self.context.subject or ''
            compiler = pybars.Compiler()
            compiled_body = compiler.compile(self.context.body)
            self_cc = True if len(recipients) == 1 else False
            for recipient in recipients:
                address = recipient.email
                data = self.recipients_data(recipient)
                if subject in self.additional.keys():
                    for nsubject, additional_data in self.additional[subject](
                            recipient, self.context.accounting_year):
                        additional_data.update(data)
                        body = format_markdown(compiled_body(additional_data))
                        msg_tag = send_mail(
                            address, nsubject, body, self.context.user,
                            attachments=self.context.attachments,
                            self_cc=self_cc)
                        SentMessageInfo.find_or_create(
                            message=self.context, tag=msg_tag, address=address)
                else:
                    body = format_markdown(compiled_body(data))
                    msg_tag = send_mail(
                        address, subject, body, self.context.user,
                        attachments=self.context.attachments,
                        self_cc=self_cc)
                    SentMessageInfo.find_or_create(
                        message=self.context, tag=msg_tag, address=address)
        sent = len(recipients)
        if sent:
            self.context.sent = datetime.now()
            self.result = {
                'status': 'success',
                'message': '%s E-Mail(s) erfolgreich versendet' % sent}
        else:
            self.result = {
                'status': 'success',
                'message': 'Keine E-Mail versendet'}
        print_recipients = [r for r in self.recipients if not r.email]
        if print_recipients:
            self.result['redirect'] = ('/api' + route_url(
                'mail_print', self.request).replace(
                    '{id}', str(self.context.id)))


@view_config(route_name='mail_duplicate', renderer='json', permission='view')
class MailDuplicateView(sw.allotmentclub.browser.base.View):

    def update(self):
        message = Message.create(
            accounting_year=datetime.now().year,
            members=self.context.members,
            externals=self.context.externals,
            subject=self.context.subject,
            body=self.context.body,
            user=self.request.user)
        for attachment in self.context.attachments:
            Attachment.create(
                message=message,
                filename=attachment.filename,
                mimetype=attachment.mimetype,
                size=attachment.size,
                data=attachment.data)
        self.result = {'status': 'success',
                       'message': 'Duplikat erstellt.'}


@view_config(route_name='mail_delete', renderer='json', permission='view')
class MailDeleteView(sw.allotmentclub.browser.base.View):

    def update(self):
        if (self.context.printed or self.context.sent):
            self.result = {
                'status': 'error',
                'message': ('Bereits gedruckte oder versendete Nachrichten '
                            'können nicht gelöscht werden.')}
            return
        self.context.delete()
        self.result = {'status': 'success', 'message': 'Nachricht gelöscht.'}


@view_config(route_name='mail_edit', renderer='json', permission='view')
class MailEditView(sw.allotmentclub.browser.base.EditJSFormView):

    title = 'Brief/E-Mail versenden'
    show_old_members = True

    def strip_html(self, body):
        if not body.startswith('<'):
            return body
        soup = BeautifulSoup(body, 'html.parser')
        texts = soup.findAll(text=True)

        def visible(element):
            if element.parent.name in ['style', 'script', '[document]',
                                       'head', 'title']:
                return False
            elif re.match('<!--.*-->', str(element)):
                return False
            return True

        result = filter(visible, texts)
        return '\n'.join(r.strip() for r in result)

    def quote_body(self, msg):
        body = self.strip_html(msg.body)
        body = '\n'.join('> {}'.format(line) for line in body.splitlines())
        recipient = (msg.members or msg.externals)[0]
        quote_header = 'Am {} schrieb {} {} <{}>:'.format(
            format_datetime(
                msg.sent, "EEEE, 'den' d. MMM Y H:m 'Uhr'", locale='de_DE'),
            recipient.firstname,
            recipient.lastname,
            recipient.email)
        return '\n\n{}\n{}'.format(quote_header, body)

    @property
    def disabled(self):
        if self.context.sent or self.context.printed:
            return True
        return False

    @property
    def load_options(self):
        collection_url = '/api/' + route_url(
            'mail_list_attachments',
            self.request).replace('{id}', str(self.context.id))
        disabled = self.disabled
        return {
            'member_ids': {
                'source': self.member_source,
                'label': 'Empfänger',
                'css_class': 'chosen',
                'disabled': disabled,
                'multiple': True,
            },
            'external_ids': {
                'source': self.external_source,
                'label': 'Externe Empfänger',
                'css_class': 'chosen',
                'disabled': disabled,
                'multiple': True,
            },
            'subject': {
                'label': 'Betreff',
                'disabled': disabled
            },
            'body': {
                'label': 'Nachricht',
                'template': 'form_markdown',
                'disabled': disabled
            },
            'attachments': {
                'label': 'Anhänge',
                'template': 'form_upload',
                'documents_collection_url': collection_url,
                'disabled': disabled
            }
        }

    def resource_data_item_title(self, item):
        meta = ('({}, {} Uhr)'.format(item.user.username, date_time(item.date))
                if item.user else '')
        return '{} {} {}'.format(
            item.filename,
            meta,
            '(Weiße Seite davor)' if item.white_page_before else '')

    @property
    def members(self):
        return super(MailEditView, self).members.filter(
            ~Member.lastname.in_(('Verein', 'Werkstatt')))

    @property
    def member_source(self):
        result = super(MailEditView, self).member_source
        verein = Member.query().filter(Member.lastname == 'Verein').one()
        result.insert(0, {u'title': u'Alle Mitglieder', u'token': verein.id})
        return result

    @property
    def external_source(self):
        return [
            {
                'token': external.id,
                'title': '{}, {} ({})'.format(
                    external.lastname,
                    external.firstname,
                    external.organization or '')
            }
            for external in (
                filter_external_recipients(
                    ExternalRecipient.query()
                ).order_by(ExternalRecipient.lastname)
            )
        ]

    @property
    def body(self):
        if self.disabled:
            return self.strip_html(self.context.body)
        if self.context.body:
            return self.context.body
        return ('Sehr geehrte{{deflection}} {{appellation}} '
                '{{title}} {{lastname}},')

    @property
    def load_data(self):
        fields = [
            ('member_ids', self.context.member_ids),
            ('external_ids', self.context.external_ids),
            ('subject', self.context.subject),
            ('body', self.body),
            ('attachments', [{'id': a.id, 'title': a.filename}
                             for a in self.context.attachments])]
        return collections.OrderedDict(fields)

    def handle_upload(self, file_):
        data = file_.file
        data.seek(0)
        data = data.read()
        return Attachment.create(
            message=self.context,
            filename=file_.filename,
            mimetype=file_.type,
            size=len(data),
            data=data), 'mail_attachment_del'


@view_config(route_name='mail_add', renderer='json', permission='view')
class MailAddView(MailEditView):

    def __init__(self, context, request):
        context = Message.create(
            user=request.user,
            accounting_year=datetime.now().year)
        context.commit()
        super(MailAddView, self).__init__(context, request)
        log_with_user(user_data_log.info, self.request.user,
                      'Nachricht %s hinzugefügt.', self.context.id)

    @property
    def route_name(self):
        return 'mail_edit'


@view_config(route_name='mail_reply', renderer='json', permission='view')
class MailReplyView(MailEditView):

    def __init__(self, context, request):
        msg = Message.create(
            user=request.user,
            in_reply_to=context,
            accounting_year=datetime.now().year)
        msg.members.extend(context.members)
        msg.externals.extend(context.externals)
        msg.body = self.quote_body(context)
        msg.subject = 'Re: {}'.format(context.subject)
        msg.commit()
        super(MailReplyView, self).__init__(msg, request)
        log_with_user(user_data_log.info, self.request.user,
                      'Nachricht %s hinzugefügt.', self.context.id)

    @property
    def route_name(self):
        return 'mail_edit'


@view_config(route_name='mail_list_attachments', renderer='json',
             permission='view')
class MailListAttachments(MailEditView):

    def update(self):
        self.result = [
            self.resource_data_item(attachment, 'mail_attachment_del')
            for attachment in self.context.attachments]


@view_config(route_name='mail_attachment_white_page', renderer='json',
             permission='view')
class AttachmentWhitePageView(MailEditView):

    def update(self):
        self.context.white_page_before = not self.context.white_page_before
        self.result = {'title': self.resource_data_item_title(self.context)}


@view_config(route_name='mail_attachment_del', renderer='json',
             permission='view')
class AttachmentDelView(sw.allotmentclub.browser.base.DeleteView):

    model = Attachment


@view_config(route_name='mail_attachment_download', renderer='json',
             permission='view')
class AttachmentDownloadView(object):

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def __call__(self):
        response = self.request.response
        response.set_cookie('fileDownload', value='true')
        response.content_type = self.context.mimetype
        response.content_length = int(self.context.size)
        response.content_disposition = (
            'attachment; filename="{}"'.format(self.context.filename))
        response.app_iter = FileIter(BytesIO(self.context.data))
        return response


@view_config(route_name='mail_postmark_open_tracking_webhook',
             permission=NO_PERMISSION_REQUIRED)
class PostmarkOpenTrackingWebhookView(object):

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def opened(self, info, data):
        received_date = dateutil.parser.parse(data['ReceivedAt'])
        native = pytz.timezone('Etc/GMT+5')
        local = pytz.timezone('Europe/Berlin')
        received = date_time(
            received_date.replace(tzinfo=native).astimezone(local))
        info.status = (
            'Geöffnet am {} Uhr in {} (IP: {}) für {} Sekunden.'.format(
                received, data['Geo'].get('City', 'Unbekannt'),
                data['Geo'].get('IP', 'Unbekannt'),
                data.get('ReadSeconds', 'Unbekannt')))

    def bounced(self, info, data):
        local = pytz.timezone('Europe/Berlin')
        bounce_date = date_time(
            dateutil.parser.parse(data['BouncedAt']).astimezone(local))
        info.status = '{} am {} Uhr: {}'.format(data['Name'], bounce_date,
                                                data['Description'])

    def delivered(self, info, data):
        local = pytz.timezone('Europe/Berlin')
        delivery_date = date_time(
            dateutil.parser.parse(data['DeliveredAt']).astimezone(local))
        info.status = 'Zugestellt am {} Uhr.'.format(delivery_date)

    def __call__(self):
        data = self.request.json
        recipient = data.get('Recipient', data.get('Email'))
        if not recipient:
            raise ValueError("No recipient found: %r" % data)
        data['Tag'] = data['Tag'].replace(' ', '')
        info = (SentMessageInfo
                .query()
                .filter(SentMessageInfo.tag == data['Tag'])
                .filter(SentMessageInfo.address == recipient)
                .one_or_none())
        if not info:
            info = (SentMessageInfo
                    .query()
                    .filter(SentMessageInfo.tag == data['Tag'])
                    .first())
            if info is None:
                return Response('ok')
            info = SentMessageInfo.create(
                organization_id=info.organization_id,
                message=info.message, tag=info.tag, address=recipient)
        if 'ReceivedAt' in data:
            self.opened(info, data)
        elif 'BouncedAt' in data:
            self.bounced(info, data)
        elif 'DeliveredAt' in data:
            self.delivered(info, data)
        else:
            raise RuntimeError(
                "Could not identify given data: %r " % data)
        return Response('ok')


@view_config(route_name='mail_postmark_inbound_webhook',
             permission=NO_PERMISSION_REQUIRED)
class PostmarkInboundWebhookView(object):

    organization_id = 1  # New messages are added on Verwaltung Roter See

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def sent_date(self, data):
        local = pytz.timezone('Europe/Berlin')
        return dateutil.parser.parse(data['Date'][:31]).astimezone(local)

    def __call__(self):
        data = self.request.json
        sender = data['FromFull']['Email'].lower()
        if sender == 'vorstand@roter-see.de':
            return Response('ok')
        message = Message.create(
            inbound=True,
            organization_id=self.organization_id,
            user=User.by_username('system'))
        member = Member.query().filter(Member.email == sender).first()
        if member:
            message.members.append(member)
            message.organization_id = member.organization_id
        else:
            external = (ExternalRecipient.query()
                        .filter(ExternalRecipient.email == sender)
                        .first())
            if not external:
                external = ExternalRecipient.create(email=sender)
                external.lastname = data['FromFull']['Name']
                external.organization_id = self.organization_id
            message.externals.append(external)
            message.organization_id = external.organization_id
        message.subject = data['Subject']
        message.body = (
            data['HtmlBody'] if data['HtmlBody'] else data['TextBody'])
        message.sent = self.sent_date(data)
        message.accounting_year = message.sent.year
        for attachment in data['Attachments']:
            Attachment.create(
                message=message,
                organization_id=message.organization_id,
                filename=attachment["Name"],
                mimetype=attachment["ContentType"],
                size=attachment["ContentLength"],
                data=base64.b64decode(attachment["Content"]))
        return Response('ok')



"""Add new data to members.

Revision ID: 4d42c583ac73
Revises: None
Create Date: 2014-10-27 13:32:02.315587

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '4d42c583ac73'
down_revision = None


def upgrade():
    op.add_column(
        'member', sa.Column('haus_nr', sa.String(length=7), nullable=True))
    op.add_column(
        'member', sa.Column('ort', sa.String(length=50), nullable=True))
    op.add_column(
        'member', sa.Column('plz', sa.String(length=6), nullable=True))
    op.add_column(
        'member', sa.Column('strasse', sa.String(length=100), nullable=True))
    op.add_column(
        'member',
        sa.Column('zaehlernummer', sa.String(length=50), nullable=True))
    op.drop_column('member', 'titel')
    op.create_unique_constraint(None, 'member', ['haus_nr'])
    op.create_unique_constraint(None, 'member', ['zaehlernummer'])


def downgrade():
    op.drop_constraint(None, 'member')
    op.drop_constraint(None, 'member')
    op.add_column(
        'member', sa.Column('titel', sa.VARCHAR(length=20), nullable=True))
    op.drop_column('member', 'zaehlernummer')
    op.drop_column('member', 'strasse')
    op.drop_column('member', 'plz')
    op.drop_column('member', 'ort')
    op.drop_column('member', 'haus_nr')



"""Remove formletters (finally)

Revision ID: 1842f7373dcd
Revises: 0593bb8c05b6
Create Date: 2016-07-05 17:23:43.155107

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = '1842f7373dcd'
down_revision = '0593bb8c05b6'


def upgrade():
    op.drop_table('formletter')


def downgrade():
    op.create_table(
        'formletter',
        sa.Column('id', sa.INTEGER(), nullable=False),
        sa.Column('kind', postgresql.ENUM(
            u'Standard', u'Abrechnung Energie', u'Abschl\xe4ge Energie',
            u'Einladung Mitgliederversammlung', u'Info-Brief',
            u'Einzugserm\xe4chtigung', u'Fehlende Arbeitsstunden',
            name='form_letter_kind'), autoincrement=False, nullable=False),
        sa.Column('mimetype', sa.VARCHAR(length=30), autoincrement=False,
                  nullable=True),
        sa.Column('size', sa.VARCHAR(length=20), autoincrement=False,
                  nullable=True),
        sa.Column('data', postgresql.BYTEA(), autoincrement=False,
                  nullable=True),
        sa.Column('date', postgresql.TIMESTAMP(), autoincrement=False,
                  nullable=False),
        sa.Column('user_id', sa.INTEGER(), autoincrement=False,
                  nullable=False),
        sa.Column('status', postgresql.ENUM(
            u'Wartet', u'In Bearbeitung', u'Fehler', u'Fertig gestellt',
            name='form_letter_status'), autoincrement=False, nullable=False),
        sa.Column('with_address', sa.BOOLEAN(), autoincrement=False,
                  nullable=True),
        sa.Column('content', sa.TEXT(), autoincrement=False, nullable=True),
        sa.Column('subject', sa.VARCHAR(), autoincrement=False, nullable=True),
        sa.ForeignKeyConstraint(
            ['user_id'], [u'user.id'], name=u'formletter_user_id_fkey'),
        sa.PrimaryKeyConstraint('id', name=u'formletter_pkey')
    )



"""re #1: Update Member model.

Revision ID: 48019c236591
Revises: 46276fff5bc2
Create Date: 2014-12-02 19:03:49.975576

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '48019c236591'
down_revision = '46276fff5bc2'


def upgrade():
    op.create_table(
        'allotment',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('number', sa.Integer(), nullable=False),
        sa.Column('member_id', sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(['member_id'], [u'member.id'], ),
        sa.PrimaryKeyConstraint('id')
    )
    op.create_table(
        'parcel',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('number', sa.Integer(), nullable=False),
        sa.Column('allotment_id', sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(['allotment_id'], [u'allotment.id'], ),
        sa.PrimaryKeyConstraint('id')
    )
    op.add_column(
        u'member',
        sa.Column('appellation', sa.String(length=20), nullable=True))
    op.add_column(
        u'member',
        sa.Column('city', sa.String(length=50), nullable=True))
    op.add_column(
        u'member',
        sa.Column('country', sa.String(length=50), nullable=True))
    op.add_column(
        u'member',
        sa.Column('firstname', sa.String(length=50), nullable=True))
    op.add_column(
        u'member',
        sa.Column('lastname', sa.String(length=50), nullable=True))
    op.add_column(
        u'member', sa.Column('mobile', sa.String(length=50), nullable=True))
    op.add_column(
        u'member', sa.Column('phone', sa.String(length=50), nullable=True))
    op.add_column(
        u'member', sa.Column('street', sa.String(length=100), nullable=True))
    op.add_column(
        u'member', sa.Column('title', sa.String(length=20), nullable=True))
    op.add_column(
        u'member', sa.Column('zip', sa.String(length=6), nullable=True))
    op.drop_column(u'member', 'ort')
    op.drop_column(u'member', 'zaehlernummer')
    op.drop_column(u'member', 'plz')
    op.drop_column(u'member', 'anrede')
    op.drop_column(u'member', 'nachname')
    op.drop_column(u'member', 'haus_nr')
    op.drop_column(u'member', 'strasse')
    op.drop_column(u'member', 'vorname')


def downgrade():
    op.add_column(
        u'member', sa.Column('vorname', sa.VARCHAR(length=50), nullable=True))
    op.add_column(
        u'member', sa.Column('strasse', sa.VARCHAR(length=100), nullable=True))
    op.add_column(
        u'member', sa.Column('haus_nr', sa.VARCHAR(length=7), nullable=True))
    op.add_column(
        u'member', sa.Column('nachname', sa.VARCHAR(length=50), nullable=True))
    op.add_column(
        u'member', sa.Column('anrede', sa.VARCHAR(length=20), nullable=True))
    op.add_column(
        u'member', sa.Column('plz', sa.VARCHAR(length=6), nullable=True))
    op.add_column(
        u'member',
        sa.Column('zaehlernummer', sa.VARCHAR(length=50), nullable=True))
    op.add_column(
        u'member', sa.Column('ort', sa.VARCHAR(length=50), nullable=True))
    op.drop_column(u'member', 'zip')
    op.drop_column(u'member', 'title')
    op.drop_column(u'member', 'street')
    op.drop_column(u'member', 'phone')
    op.drop_column(u'member', 'mobile')
    op.drop_column(u'member', 'lastname')
    op.drop_column(u'member', 'firstname')
    op.drop_column(u'member', 'country')
    op.drop_column(u'member', 'city')
    op.drop_column(u'member', 'appellation')
    op.drop_table('parcel')
    op.drop_table('allotment')



"""Add table for Grundsteuer B.

Revision ID: 9dadb9f9561f
Revises: b7de0b4c2bb6
Create Date: 2016-07-03 16:41:55.773497

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '9dadb9f9561f'
down_revision = 'b7de0b4c2bb6'


def upgrade():
    op.create_table(
        'grundsteuerb',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('parcel_id', sa.Integer(), nullable=False),
        sa.Column('value', sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(['parcel_id'], [u'parcel.id'], ),
        sa.PrimaryKeyConstraint('id')
    )
    op.execute(
        "INSERT into grundsteuerb (parcel_id, value) VALUES (175, 39800);")
    op.execute(
        "INSERT into grundsteuerb (parcel_id, value) VALUES (172, 39800);")
    op.execute(
        "INSERT into grundsteuerb (parcel_id, value) VALUES (147, 39800);")
    op.execute(
        "INSERT into grundsteuerb (parcel_id, value) VALUES (128, 39800);")
    op.execute(
        "INSERT into grundsteuerb (parcel_id, value) VALUES (119, 39800);")
    op.execute(
        "INSERT into grundsteuerb (parcel_id, value) VALUES (86, 39800);")
    op.execute(
        "INSERT into grundsteuerb (parcel_id, value) VALUES (87, 79600);")
    op.execute(
        "INSERT into grundsteuerb (parcel_id, value) VALUES (85, 59700);")
    op.execute(
        "INSERT into grundsteuerb (parcel_id, value) VALUES (177, 59700);")


def downgrade():
    op.drop_table('grundsteuerb')



"""re #11: Allow adding documents to member

Revision ID: c4dbe65a7810
Revises: 3efd1e093b36
Create Date: 2021-01-05 11:23:35.604257

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = 'c4dbe65a7810'
down_revision = '3efd1e093b36'


def upgrade():
    op.create_table(
        'memberattachment',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('organization_id', sa.Integer(), nullable=False),
        sa.Column('member_id', sa.Integer(), nullable=False),
        sa.Column('name', sa.String(length=100), nullable=True),
        sa.Column('mimetype', sa.String(length=30), nullable=True),
        sa.Column('size', sa.String(length=20), nullable=True),
        sa.Column('data', sa.LargeBinary(length=10485760), nullable=True),
        sa.ForeignKeyConstraint(['member_id'], ['member.id'], ),
        sa.PrimaryKeyConstraint('id')
    )


def downgrade():
    op.drop_table('memberattachment')



"""Add banking accounts to booking kinds.

Revision ID: 388a201feabf
Revises: 5fe4feac157
Create Date: 2015-04-01 13:17:35.841429

"""

# revision identifiers, used by Alembic.
from alembic import op
import sqlalchemy as sa

revision = '388a201feabf'
down_revision = '5fe4feac157'


def upgrade():
    op.add_column(
        'bookingkind',
        sa.Column('banking_account_id', sa.Integer(), nullable=True))
    op.execute("UPDATE bookingkind SET banking_account_id = 2 WHERE id = 1;")
    op.execute("UPDATE bookingkind SET banking_account_id = 3 WHERE id = 2;")
    op.execute("UPDATE bookingkind SET banking_account_id = 4 WHERE id = 3;")
    op.execute("UPDATE bookingkind SET banking_account_id = 4 WHERE id = 4;")
    op.execute("UPDATE bookingkind SET banking_account_id = 4 WHERE id = 5;")


def downgrade():
    op.drop_column('bookingkind', 'banking_account_id')



"""Correct allotments and parcels.

Revision ID: 24a3f3740664
Revises: 186285f37c74
Create Date: 2015-03-05 10:34:51.011115

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = '24a3f3740664'
down_revision = '186285f37c74'


def upgrade():
    op.execute("UPDATE parcel SET number = 103 WHERE id = 119;")
    op.execute("INSERT into parcel (number, allotment_id) VALUES (131, 185);")
    op.execute("UPDATE parcel SET number = 135 WHERE id = 159;")
    op.execute("UPDATE parcel SET number = 136 WHERE id = 179;")
    op.execute("UPDATE parcel SET number = 123 WHERE id = 197;")
    op.execute("UPDATE parcel SET number = 122 WHERE id = 177;")


def downgrade():
    pass



"""Add accounting_year to inbound messages.

Revision ID: ff18c87afd8e
Revises: 88942e83a048
Create Date: 2017-05-26 11:27:47.675422

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = 'ff18c87afd8e'
down_revision = '88942e83a048'


def upgrade():
    op.execute(
        """UPDATE message
           SET accounting_year = 2017
           WHERE inbound = true;""")


def downgrade():
    pass



# -*- coding: utf-8 -*-
from .model import Object, MAX_FILE_SIZE
from sqlalchemy import Column, Integer, String, ForeignKey, LargeBinary
from sqlalchemy import DateTime
import sqlalchemy.orm


def get_current_user(*args, **kw):
    import pyramid
    user = pyramid.threadlocal.get_current_request().user
    if user:
        return user.id


class Depot(Object):
    """ Die Ablage."""

    name = Column(String(100), default=u'')
    mimetype = Column(String(100), default=u'')
    size = Column(String(20), default=u'')
    data = Column(LargeBinary(MAX_FILE_SIZE))
    date = Column(DateTime, nullable=False)
    user_id = Column(Integer, ForeignKey('user.id'), nullable=False,
                     default=get_current_user)
    user = sqlalchemy.orm.relation(
        'User', uselist=False, backref='depot_items')



"""Add direct debit date.

Revision ID: 411ca2f55194
Revises: 2b1aa8adb13b
Create Date: 2015-03-30 09:28:12.197684

"""

# revision identifiers, used by Alembic.
from alembic import op
import sqlalchemy as sa

revision = '411ca2f55194'
down_revision = '2b1aa8adb13b'


def upgrade():
    op.add_column('member',
                  sa.Column('direct_debit_date', sa.DateTime(), nullable=True))


def downgrade():
    op.drop_column('member', 'direct_debit_date')



# encoding=utf-8
"""Add accounting year to bookings.

Revision ID: 51aecbad82b3
Revises: 43123080ac3a
Create Date: 2015-03-11 16:27:13.098704

"""

# revision identifiers, used by Alembic.
from __future__ import unicode_literals
from alembic import op
import sqlalchemy as sa

revision = '51aecbad82b3'
down_revision = '43123080ac3a'


def upgrade():
    op.add_column(
        'booking',
        sa.Column('accounting_year', sa.Integer(), nullable=True))
    op.execute(
        "UPDATE booking SET accounting_year = EXTRACT(YEAR FROM booking_day);")
    op.execute(
        """INSERT into bankingaccount (number, name)
           VALUES ('1', 'Mitgliedsbeiträge')""")
    op.execute(
        """INSERT into bankingaccount (number, name)
           VALUES ('2', 'Fehlende Arbeitseinsätze')""")
    op.execute(
        """INSERT into bankingaccount (number, name)
           VALUES ('3', 'Abschläge Energie')""")


def downgrade():
    op.drop_column('booking', 'accounting_year')
    op.execute("DELETE FROM bankingaccount WHERE id != 1")



"""Add iban and bic to account bookings.

Revision ID: 41e271c4e3e8
Revises: 411ca2f55194
Create Date: 2015-03-30 18:59:16.733989

"""

# revision identifiers, used by Alembic.
from alembic import op
import sqlalchemy as sa

revision = '41e271c4e3e8'
down_revision = '411ca2f55194'


def upgrade():
    op.add_column(
        'booking', sa.Column('bic', sa.String(length=11), nullable=True))
    op.add_column(
        'booking', sa.Column('iban', sa.String(length=34), nullable=True))
    op.drop_column('booking', 'account_number')


def downgrade():
    op.add_column(
        'booking',
        sa.Column('account_number', sa.VARCHAR(length=56), nullable=True))
    op.drop_column('booking', 'iban')
    op.drop_column('booking', 'bic')
